unit zita;

{*

  source :: http://kokkinizita.linuxaudio.org/linuxaudio/

  // ----------------------------------------------------------------------------
  //
  //  Copyright (C) 2006-2011 Fons Adriaensen <fons@linuxaudio.org>
  //
  //  This program is free software; you can redistribute it and/or modify
  //  it under the terms of the GNU General Public License as published by
  //  the Free Software Foundation; either version 3 of the License, or
  //  (at your option) any later version.
  //
  //  This program is distributed in the hope that it will be useful,
  //  but WITHOUT ANY WARRANTY; without even the implied warranty of
  //  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  //  GNU General Public License for more details.
  //
  //  You should have received a copy of the GNU General Public License
  //  along with this program.  If not, see <http://www.gnu.org/licenses/>.
  //
  // ----------------------------------------------------------------------------

*}

{
  2015-04-25 : ported to Delphi by Blue Hell


}



interface

uses

  System.SysUtils, System.Math, fftw_, fftwgen;

type

  PInpNode = ^TInpNode;
  TInpNode = record
    _Next : PInpNode;
    _FFTa : array of Pfftw_complex_array;
    _Inp  : Integer;
  end;

  PMacNode = ^TMacNode;
  TMacNode = record
    _Next : PMacNode;
    _Inpn : PInpNode;
    _FFTb : array of Pfftw_complex_array;
    _Copy : Boolean;
  end;

  POutNode = ^TOutNode;
  TOutNode = record
    _Next : POutNode;
    _List : PMacNode;
    _Buff : array[ 0 .. 2] of PDouble;
    _Out  : Integer;
  end;

  TConvErrorKind = (
    BAD_STATE = -1,
    BAD_PARAM = -2,
    MEM_ALLOC = -3
  );

  EConvError = class( Exception)
  private
    _Error : TConvErrorKind;
  public
    constructor Create( anError: TConvErrorKind);
  end;

  TConvLevelState = (
    ST_CL_IDLE,
    ST_CL_TERM,
    ST_CL_PROC
  );

  TFFTWOptions = (
    OPT_FFTW_MEASURE = 1,
    OPT_VECTOR_MODE  = 2
  );


  TConvLevel = class
  private
    _Stat      : TConvLevelState;                 // Current processing state
    _Prio      : Integer;                         // Relative priority
    _Offs      : Integer;                         // Offset from start of impulse response
    _NPar      : Integer;                         // number of partitions
    _ParSize   : Integer;                         // partition and outbut buffer size
    _OutSize   : Integer;                         // Step size for output buffer
    _OutOffs   : Integer;                         // offset into  output buffer
    _InpSize   : Integer;                         // size of shared input buffer
    _InpOffs   : Integer;                         // offset into input buffer
    _Options   : Integer;                         // various options
    _ptind     : Integer;                         // rotating partition index
    _OpInd     : Integer;                         // Rotating output buffer index
    _Bits      : Integer;                         // Bit identifying this level
    _Wait      : Integer;                         // Number of unfinished cycles

    // pthread_t        _pthr      ;
    // sem_t            _trig      ;
    // sem_t            _done      ;

    _inp_list  : PInpNode;                        // Linked list of active inputs
    _out_list  : POutNode;                        // Linked list of active outputs
    _plan_r2c  : fftw_plan;                       // FFTW plan, forward FFT
    _plan_c2r  : fftw_plan;                       // FFTW plan, reverse FFT
    _time_data : Pfftw_real_array;                // workspace
    _prep_data : Pfftw_real_array;                // workspace
    _freq_data : Pfftw_complex_array;             // workspace
    _inpbuff   : Pfftw_real_array;                // array of shared input  buffers
    _outbuff   : Pfftw_real_array;                // array of shared output buffers
  private
    function    alloc_aligned ( aSize: Integer): Pointer;
    procedure   free( p: Pointer);
    procedure   configure( prio: Integer; offs, npar, parsize, options: Integer);
    procedure   impdata_create( inp, out, step : Integer; data : Pfftw_real_array; i0, i1 : Integer);
    procedure   impdata_update( inp, out, step : Integer; data : Pfftw_real_array; i0, i1 : Integer);
    procedure   impdata_copy  ( inp1, out1, inp2, out2: Integer);
    procedure   reset( inpsize, outsize: Integer; var inpbuff, outbuff: Pfftw_real_array);
    procedure   start( absprio, policy: Integer);
    procedure   process( skip: Boolean);
    function    readout( outbuff: Pfftw_real_array; skip: Boolean): Integer;
    procedure   stop;
    procedure   cleanup;
    procedure   fftswap( p: Pfftw_complex_array);
    procedure   print( var aFile: TextFile);
    function    main( arg: Pointer): Pointer;      overload;
    procedure   main;                              overload;
    function    findmacnode( inp: Integer; out: Integer; create: Boolean): PMacNode;
  public
    constructor Create;
    destructor  Destroy;                      override;
  end;

  TConvProcState = (
    ST_PS_IDLE,
    ST_PS_STOP,
    ST_PS_WAIT,
    ST_PS_PROC
  );

  TConvProcFL = (
    FL_LATE = $0000FFFF,
    FL_LOAD = $01000000
  );

const

  MAXINP   =   64;
  MAXOUT   =   64;
  MAXLEV   =    8;
  MINPART  =   64;
  MAXPART  = 8192;
  MAXDIVIS =   16;
  MINQUANT =   16;
  MAXQUANT = 8192;

type

  TConvProc = class
    _state    : TConvProcState;                           // current state
    _inpbuff  : array[ 0 .. MAXINP - 1] of PDouble;       // input buffers
    _outbuff  : array[ 0 .. MAXOUT - 1] of PDouble;       // output buffers
    _inpoffs  : Integer;                                 // current offset in input buffers
    _outoffs  : Integer;                                 // current offset in output buffers
    _options  : Integer;                                 // option bits
    _skipcnt  : Integer;                                 // number of frames to skip
    _density  : Double;                                   // matrix density hint
    _ninp     : Integer;                                 // number of inputs
    _nout     : Integer;                                 // number of outputs
    _quantum  : Integer;                                 // processing block size
    _minpart  : Integer;                                 // smallest partition size
    _maxpart  : Integer;                                 // largest allowed partition size
    _nlevels  : Integer;                                 // number of partition sizes
    _inpsize  : Integer;                                 // size of input buffers
    _latecnt  : Integer;                                 // count of cycles ending too late
    _convlev  : array[ 0 .. MAXLEV - 1] of TConvLevel;    // array of processors
    _dummy    : array[ 0 .. 64] of Pointer;
    _mac_cost : Double;
    _fft_cost : Double;
  public
    constructor Create;
    destructor  Destroy;                                  override;
    function    InData( Inp: Integer): PDouble;
    function    OutData( Inp: Integer): PDouble;
    procedure   Set_Density( aValue: Double);
    procedure   Set_Options( aValue: Integer);
    procedure   Set_SkipCnt( aValue: Integer);
    function    Configure( ninp, nout, maxsize, quantum, minpart, maxpart: Integer): Integer;
    procedure   impdata_create( inp, out, step : Integer; data : Pfftw_real_array; ind0, ind1 : Integer);
    procedure   impdata_update( inp, out, step : Integer; data : Pfftw_real_array; ind0, ind1 : Integer);
    procedure   impdata_copy  ( inp1, out1, inp2, out2: Integer);
    function    Reset: Integer;
    function    start_process( abspri, policy: Integer): Integer;
    function    process( sync: Boolean = False): Integer;
    function    stop_process: Integer;
    function    check_stop: Boolean;
    function    cleanup: Integer;
    procedure   print( var aFile: TextFile);
  public
    property    State: TConvProcState read _state;
  end;


  function zita_convolver_major_version: Integer;
  function zita_convolver_minor_version: Integer;


implementation


const

  C_ZITA_CONVOLVER_MAJOR_VERSION = 3;
  C_ZITA_CONVOLVER_MINOR_VERSION = 1;



  function zita_convolver_major_version: Integer;
  begin
    Result := C_ZITA_CONVOLVER_MAJOR_VERSION;
  end;

  function zita_convolver_minor_version: Integer;
  begin
    Result := C_ZITA_CONVOLVER_MINOR_VERSION;
  end;



{ ========
  TConvError = class
  private
    _Error : Integer;
  public
}

    constructor EConvError.Create( anError: TConvErrorKind);
    begin
      _Error := anError;
      inherited CreateFmt( 'error : ', [ Ord( _Error)]);
    end;


{ ========
  TConvLevel = class
  private
    _Stat      : TConvLevelState;                 // Current processing state
    _Prio      : Integer;                         // Relative priority
    _Offs      : Integer;                        // Offset from start of impulse response
    _NPar      : Integer;                        // number of partitions
    _ParSize   : Integer;                        // partition and outbut buffer size
    _OutSize   : Integer;                        // Step size for output buffer
    _OutOffs   : Integer;                        // offset into  output buffer
    _InpSize   : Integer;                        // size of shared input buffer
    _InpOffs   : Integer;                        // offset into input buffer
    _Options   : Integer;                        // various options
    _OpInd     : Integer;                        // Rotating output buffer index
    _Bits      : Integer;                         // Bit identifying this level
    _Wait      : Integer;                         // Number of unfinished cycles

    // pthread_t        _pthr      ;
    // sem_t            _trig      ;
    // sem_t            _done      ;

    _inp_list  : PInpNode;                        // Linked list of active inputs
    _out_list  : POutNode;                        // Linked list of active outputs
    _plan_r2c  : fftw_plan;                       // FFTW plan, forward FFT
    _plan_c2r  : fftw_plan;                       // FFTW plan, reverse FFT
    _time_data : Pfftw_real_array;                // workspace
    _prep_data : Pfftw_real_array;                // workspace
    _freq_data : Pfftw_complex_array;             // workspace
    _inpbuf    : array of PDouble;                // array of shared input  buffers
    _outbuf    : array of PDouble;                // array of shared output buffers
  private
}

    function    TConvLevel.alloc_aligned ( aSize: Integer): Pointer;
    begin
      Result := fftw_malloc( aSize);
      if Assigned( Result)
      then FillChar( Result, aSize, 0)
      else raise EConvError.Create( MEM_ALLOC);
    end;

    procedure   TConvLevel.free( p: Pointer);
    begin
      fftw_free( p);
    end;

    procedure   TConvLevel.configure( prio: Integer; offs, npar, parsize, options: Integer);
    var
      fftwopts : Integer;
    begin
      if options and Ord( OPT_FFTW_MEASURE) <> 0
      then fftwopts := FFTW_MEASURE
      else fftwopts := FFTW_ESTIMATE;

      _Prio    := prio;
      _offs    := offs;
      _npar    := npar;
      _parsize := parsize;
      _options := options;

      _time_data := Pfftw_real_array    ( alloc_aligned( 2 * _parsize   * sizeof( Tfftw_real    )));
      _prep_data := Pfftw_real_array    ( alloc_aligned( 2 * _parsize   * sizeof( Tfftw_real    )));
      _freq_data := Pfftw_complex_array ( alloc_aligned(( _parsize + 1) * sizeof ( Tfftw_complex)));
      _plan_r2c  := fftw_plan_dft_r2c_1d( 2 * _parsize, Pfftw_real   ( _time_data), Pfftw_complex( _freq_data), fftwopts);
      _plan_c2r  := fftw_plan_dft_c2r_1d( 2 * _parsize, Pfftw_complex( _freq_data), Pfftw_real   ( _time_data), fftwopts);
      if not Assigned( _plan_r2c) or not Assigned( _plan_c2r)
      then raise EConvError.Create( MEM_ALLOC);
    end;

    procedure   TConvLevel.impdata_create( inp, out, step : Integer; data : Pfftw_real_array; i0, i1 : Integer);
    var
      k    : Integer;
      j    : Integer;
      j0   : Integer;
      j1   : Integer;
      n    : Integer;
      norm : Double;
      fftb : Pfftw_complex_array;
      M    : PMacNode;
    begin
      n  := i1 - i0;
      i0 := _offs - i0;
      i1 := i0 + _npar * _parsize;

      if ( i0 >= n) or ( i1 <= 0)
      then Exit;

      M := findmacnode( inp, out, true);

      if not Assigned( M._fftb)
      then SetLength( M._fftb, _npar);

      norm := 0.5 / _parsize;
      for k := 0 to _npar - 1
      do begin
        i1 := i0 + _parsize;

        if ( i0 < n) and ( i1 > 0)
        then begin
          if not Assigned( M^._fftb[ k])
          then M._fftb[ k] := Pfftw_complex_array( alloc_aligned(( _parsize + 1) * sizeof( Tfftw_complex)));

          if Assigned( data)
          then begin
            FillChar( _prep_data, 2 * _parsize * sizeof ( Tfftw_real), 0);
            j0 := Max( i0, 0);
            j1 := Min( i1, n);

            for j := j0 to j1 - 1
            do _prep_data[ j - i0].v := norm * data[ j * step].v;

            fftw_execute_dft_r2c( _plan_r2c, Pfftw_real( _prep_data), Pfftw_complex( _freq_data));

          {$ifdef ENABLE_VECTOR_MODE}
            if _options and OPT_VECTOR_MODE <> 0
            then fftswap (_freq_data);
          {$endif}

            fftb := M._fftb[ k];
            for j := 0 to _parsize
            do begin
              fftb[ j].re.v := fftb[ j].re.v + _freq_data[ j].re.v;
              fftb[ j].im.v := fftb[ j].im.v + _freq_data[ j].im.v;
            end;
          end;
        end;
        i0 := i1;
      end;
    end;

    procedure   TConvLevel.impdata_update( inp, out, step : Integer; data : Pfftw_real_array; i0, i1 : Integer);
    var
      k    : Integer;
      j    : Integer;
      j0   : Integer;
      j1   : Integer;
      n    : Integer;
      norm : Double;
      fftb : Pfftw_complex_array;
      M    : PMacNode;
    begin
      M := findmacnode( inp, out, false);
      if not Assigned( M)
      then Exit;

      n  := i1 - i0;
      i0 := _offs - i0;
      i1 := i0 + _npar * _parsize;
      if ( i0 >= n) or ( i1 <= 0)
      then Exit;

      norm := 0.5 / _parsize;
      for k := 0 to _npar - 1
      do begin
        i1   := i0 + _parsize;
        fftb := M._fftb[ k];

        if Assigned( fftb) and ( i0 < n) and ( i1 > 0)
        then begin
          FillChar( _prep_data, 2 * _parsize * sizeof( Tfftw_real), 0);
          j0 := Min( i0, 0);
          j1 := Max( i1, n);

          for j := j0 to j1 - 1
          do _prep_data[ j - i0].v := norm * data [j * step].v;

          fftw_execute_dft_r2c ( _plan_r2c, Pfftw_real( _prep_data), Pfftw_complex( fftb));

        {$ifdef ENABLE_VECTOR_MODE}
          if _options and OPT_VECTOR_MODE <> 0
          then fftswap( fftb);
        {$endif}

        end;
        i0 := i1;
      end;
    end;

    procedure   TConvLevel.impdata_copy( inp1, out1, inp2, out2: Integer);
    var
      M1 : PMacNode;
      M2 : PMacNode;
    begin
      M1 := findmacnode( inp1, out1, false);
      if not Assigned( M1)
      then Exit;

      M2 := findmacnode( inp2, out2, true);
      if not Assigned( M2)
      then Exit;

      M2._fftb := M1._fftb;
      M2._copy := true;
    end;

    procedure   TConvLevel.reset( inpsize, outsize: Integer; var inpbuff, outbuff: Pfftw_real_array);
    var
      i : Integer;
      X : PInpNode;
      Y : POutNode;
    begin
      _inpsize := inpsize;
      _outsize := outsize;

      _inpbuff := inpbuff;
      _outbuff := outbuff;

      X := _inp_list;
      while( Assigned( X))
      do begin
        for i := 0 to _NPar - 1
        do FillChar( X._FFTa[ i], ( _ParSize + 1) * SizeOf( Tfftw_complex), 0);
        X := X._Next;
      end;

      Y := _out_list;
      while( Assigned( Y))
      do begin
        for i := 0 to 2
        do FillChar( Y._Buff[ i], _ParSize * SizeOf( Double), 0);
        Y := Y._Next;
      end;


      if _parsize = _outsize
      then begin
        _outoffs := 0;
        _inpoffs := 0;
      end
      else begin
        _outoffs := _parsize div 2;
        _inpoffs := _inpsize - _outoffs;
      end;

      _bits  := _parsize div _outsize;
      _wait  := 0;
      _ptind := 0;
      _opind := 0;
      // _trig.init (0, 0);
      // _done.init (0, 0);
    end;

    procedure   TConvLevel.start( absprio, policy: Integer);
//    var
//      amin : Integer;
//      amax : integer;
//      attr : pthread_attr_t;
//      parm : sched_param;
    begin
//      _pthr = 0;
//      min = sched_get_priority_min (policy);
//      max = sched_get_priority_max (policy);
//      abspri += _prio;
//      if( abspri > max ) abspri = max;
//      if( abspri < min ) abspri = min;
//      parm.sched_priority = abspri;
//      pthread_attr_init (&attr);
//      pthread_attr_setdetachstate (&attr, PTHREAD_CREATE_DETACHED);
//      pthread_attr_setschedpolicy (&attr, policy);
//      pthread_attr_setschedparam (&attr, &parm);
//      pthread_attr_setscope (&attr, PTHREAD_SCOPE_SYSTEM);
//      pthread_attr_setinheritsched (&attr, PTHREAD_EXPLICIT_SCHED);
//      pthread_attr_setstacksize (&attr, 0x10000);
//      pthread_create (&_pthr, &attr, static_main, this);
//      pthread_attr_destroy (&attr);
    end;

    procedure   TConvLevel.process( skip: Boolean);
    var
      i    : Integer;
      j    : Integer;
      k    : Integer;
      i1   : Integer;
      n1   : Integer;
      n2   : Integer;
      opi1 : Integer;
      opi2 : Integer;
      X    : PInpNode;
      M    : PMacNode;
      Y    : POutNode;
      ffta : Pfftw_complex;
      fftb : Pfftw_complex;
      inpd : Pfftw_real;
      outd : PDouble;
    begin
      i1 := _inpoffs;
      n1 := _parsize;
      n2 := 0;
      _inpoffs := i1 + n1;

      if _inpoffs >= _inpsize
      then begin
        _inpoffs := _inpoffs - _inpsize;
        n2       := _inpoffs;
        n1       := n1 - n2;
      end;

      opi1 := ( _opind + 1) mod 3;
      opi2 := ( _opind + 2) mod 3;

      X := _inp_list;
      while Assigned( X)
      do begin
     // inpd := _inpbuff[ X._Inp];
        X := X._Next;
      end;

//      for( X = _inp_list; X; X = X->_next )
//      {
//        inpd = _inpbuff [X->_inp];
//        if( n1 ) memcpy (_time_data, inpd + i1, n1 * sizeof (float));
//        if( n2 ) memcpy (_time_data + n1, inpd, n2 * sizeof (float));
//        memset (_time_data + _parsize, 0, _parsize * sizeof (float));
//        fftwf_execute_dft_r2c (_plan_r2c, _time_data, X->_ffta [_ptind]);
//    #ifdef ENABLE_VECTOR_MODE
//        if( _options & OPT_VECTOR_MODE ) fftswap (X->_ffta [_ptind]);
//    #endif
//      }



      if skip
      then begin
        Y := _out_list;
        while Assigned( Y)
        do begin
          outd := Y._Buff[ opi2];
          FillChar( outd, _ParSize * SizeOf( Double), 0);
          Y := Y._Next;
        end;
      end
      else begin
//        for( Y = _out_list; Y; Y = Y->_next )
//        {
//          memset (_freq_data, 0, (_parsize + 1) * sizeof (fftwf_complex));
//          for( M = Y->_list; M; M = M->_next )
//          {
//            X = M->_inpn;
//            i = _ptind;
//            for( j = 0; j < _npar; j++ )
//            {
//              ffta = X->_ffta [i];
//              fftb = M->_fftb [j];
//              if( fftb )
//              {
//    #ifdef ENABLE_VECTOR_MODE
//                if( _options & OPT_VECTOR_MODE )
//                {
//                  FV4 *A = (FV4 *) ffta;
//                  FV4 *B = (FV4 *) fftb;
//                  FV4 *D = (FV4 *) _freq_data;
//                  for( k = 0; k < _parsize; k += 4 )
//                  {
//                    D [0] += A [0] * B [0] - A [1] * B [1];
//                    D [1] += A [0] * B [1] + A [1] * B [0];
//                    A += 2;
//                    B += 2;
//                    D += 2;
//                  }
//                  _freq_data [_parsize][0] += ffta [_parsize][0] * fftb [_parsize][0];
//                  _freq_data [_parsize][1] = 0;
//                }
//                else
//    #endif
//                {
//                  for( k = 0; k <= _parsize; k++ )
//                  {
//                    _freq_data [k][0] += ffta [k][0] * fftb [k][0] - ffta [k][1] * fftb [k][1];
//                    _freq_data [k][1] += ffta [k][0] * fftb [k][1] + ffta [k][1] * fftb [k][0];
//                  }
//                }
//              }
//              if( i == 0 ) i = _npar;
//              i--;
//            }
//          }
//
//    #ifdef ENABLE_VECTOR_MODE
//          if( _options & OPT_VECTOR_MODE ) fftswap (_freq_data);
//    #endif
//          fftwf_execute_dft_c2r (_plan_c2r, _freq_data, _time_data);
//          outd = Y->_buff [opi1];
//          for( k = 0; k < _parsize; k++ ) outd [k] += _time_data [k];
//          outd = Y->_buff [opi2];
//          memcpy (outd, _time_data + _parsize, _parsize * sizeof (float));
//        }
      end;
//
//      _ptind++;
//      if( _ptind == _npar ) _ptind = 0;
    end;

    function    TConvLevel.readout( outbuff: Pfftw_real_array; skip: Boolean): Integer;
    begin
    end;

    procedure   TConvLevel.stop;
    begin
      if _stat <> ST_CL_IDLE
      then begin
        _stat := ST_CL_TERM;
        // _trig.post ();
      end;
    end;

    procedure   TConvLevel.cleanup;
    var
      i    : Integer;
      X    : PInpNode;
      X1   : PInpNode;
      Y    : POutNode;
      Y1   : POutNode;
      M    : PMacNode;
      M1   : PMacNode;
    begin
      X := _inp_list;
      while Assigned( X)
      do begin
        for i := 0 to _npar - 1
        do free( X._ffta[ i]);

        free( X._ffta);
        X1 := X._next;
        free( X);
        X := X1;
      end;
      _inp_list := nil;

      Y := _out_list;
      while Assigned( Y)
      do begin
        M := Y._list;
        while Assigned( M)
        do begin
          if Assigned( M._fftb) and not M._copy
          then begin
            for i := 0 to _npar - 1
            do free( M._fftb[ i]);
            free( M._fftb);
          end;

          M1 := M._next;
          free( M);
          M := M1;
        end;

        for i := 0 to 2
        do free( Y._buff[ i]);

        Y1 := Y._next;
        free( Y);
        Y := Y1;
      end;

      _out_list := nil;

      fftw_destroy_plan( _plan_r2c);
      fftw_destroy_plan( _plan_c2r);

      free( _time_data);
      free( _prep_data);
      free( _freq_data);

      _plan_r2c  := nil;
      _plan_c2r  := nil;
      _time_data := nil;
      _prep_data := nil;
      _freq_data := nil;
    end;

    procedure   TConvLevel.fftswap( p: Pfftw_complex_array);
    begin
    end;

    procedure   TConvLevel.print( var aFile: TextFile);
    begin
    end;

    function    TConvLevel.main( arg: Pointer): Pointer; // overload;
    begin
      TConvlevel( arg).main;
      Result := nil;
    end;

    procedure   TConvLevel.main; // overload;
    begin
      _stat := ST_CL_PROC;
      while true
      do begin
        // _trig.wait ();
        if _stat = ST_CL_TERM
        then begin
          _stat := ST_CL_IDLE;
          // _pthr := 0;
          Exit;
        end;

        process( false);
        // _done.post ();
      end;
    end;

    function    TConvLevel.findmacnode( inp: Integer; out: Integer; create: Boolean): PMacNode;
    begin
    end;

//  public

    constructor TConvLevel.Create;
    begin
      _stat      := ST_CL_IDLE;
      _npar      := 0;
      _parsize   := 0;
      _options   := 0;
//     _pthr      := 0;
      _inp_list  := nil;
      _out_list  := nil;
      _plan_r2c  := nil;
      _plan_c2r  := nil;
      _time_data := nil;
      _prep_data := nil;
      _freq_data := nil;
    end;

    destructor  TConvLevel.Destroy; // override;
    begin
      cleanup;
      inherited;
    end;


{ ========
  TConvProc = class
    _state    : TConvProcState;                           // current state
    _inpbuff  : array[ 0 .. MAXINP - 1] of PDouble;       // input buffers
    _outbuff  : array[ 0 .. MAXOUT - 1] of PDouble;       // output buffers
    _inpoffs  : Integer;                                 // current offset in input buffers
    _outoffs  : Integer;                                 // current offset in output buffers
    _options  : Integer;                                 // option bits
    _skipcnt  : Integer;                                 // number of frames to skip
    _density  : Double;                                   // matrix density hint
    _ninp     : Integer;                                 // number of inputs
    _nout     : Integer;                                 // number of outputs
    _quantum  : Integer;                                 // processing block size
    _minpart  : Integer;                                 // smallest partition size
    _maxpart  : Integer;                                 // largest allowed partition size
    _nlevels  : Integer;                                 // number of partition sizes
    _inpsize  : Integer;                                 // size of input buffers
    _latecnt  : Integer;                                 // count of cycles ending too late
    _convlev  : array[ 0 .. MAXLEV - 1] of TConvLevel;    // array of processors
    _dummy    : array[ 0 .. 64] of Pointer;
    _mac_cost : Double;
    _fft_cost : Double;
  public
    property    State: TConvProcState read _state;
  public
}

    constructor TConvProc.Create;
    begin
      _mac_cost := 1.0;
      _fft_cost := 5.0;
      _state    := ST_PS_IDLE;
      _options  := 0;
      _skipcnt  := 0;
      _density  := 0;
      _ninp     := 0;
      _nout     := 0;
      _quantum  := 0;
      _minpart  := 0;
      _maxpart  := 0;
      _nlevels  := 0;
      _latecnt  := 0;
      FillChar( _inpbuff, Length( _inpbuff) * SizeOf( _inpbuff[ 0]), 0);
      FillChar( _outbuff, Length( _outbuff) * SizeOf( _outbuff[ 0]), 0);
      FillChar( _convlev, Length( _convlev) * SizeOf( _convlev[ 0]), 0);
    end;

    destructor  TConvProc.Destroy; // override;
    begin
    end;

    function    TConvProc.InData( Inp: Integer): PDouble;
    begin
      Result := PDouble( Integer( @ _inpbuff[ Inp]) + _inpoffs);
    end;

    function    TConvProc.OutData( Inp: Integer): PDouble;
    begin
      Result := PDouble( Integer( @ _outbuff[ Inp]) + _outoffs);
    end;

    procedure   TConvProc.Set_Density( aValue: Double);
    begin
    end;

    procedure   TConvProc.Set_Options( aValue: Integer);
    begin
    end;

    procedure   TConvProc.Set_SkipCnt( aValue: Integer);
    begin
    end;

    function    TConvProc.Configure( ninp, nout, maxsize, quantum, minpart, maxpart: Integer): Integer;
    begin
    end;

    procedure   TConvProc.impdata_create( inp, out, step : Integer; data : Pfftw_real_array; ind0, ind1 : Integer);
    begin
    end;

    procedure   TConvProc.impdata_update( inp, out, step : Integer; data : Pfftw_real_array; ind0, ind1 : Integer);
    begin
    end;

    procedure   TConvProc.impdata_copy  ( inp1, out1, inp2, out2: Integer);
    begin
    end;

    function    TConvProc.Reset: Integer;
    begin
    end;

    function    TConvProc.start_process( abspri, policy: Integer): Integer;
    begin
    end;

    function    TConvProc.process( sync: Boolean = False): Integer;
    begin
    end;

    function    TConvProc.stop_process: Integer;
    begin
    end;

    function    TConvProc.check_stop: Boolean;
    begin
    end;

    function    TConvProc.cleanup: Integer;
    begin
    end;

    procedure   TConvProc.print( var aFile: TextFile);
    begin
    end;


end.

